import sqlalchemy
from fakeurldetector import Config, LOGGER
metadata = sqlalchemy.MetaData()


users = sqlalchemy.Table(
    "users",
    metadata,
    sqlalchemy.Column("uuid", sqlalchemy.String(40), primary_key=True),
    sqlalchemy.Column("name", sqlalchemy.String(30)),
    sqlalchemy.Column("client", sqlalchemy.String(10)),
    sqlalchemy.Column("bot_token", sqlalchemy.String(40)),
    sqlalchemy.Column("telegram_user_id", sqlalchemy.String(20))
)

engine = sqlalchemy.create_engine(Config.DB_URL, connect_args={"check_same_thread": False})
metadata.create_all(engine)

LOGGER.info("Created Tables and ORM Engine")



from typing import List, Optional, Annotated, Any, Dict
from uuid import UUID

from pydantic import BaseModel, AfterValidator, model_validator, HttpUrl


def check_client(client: str):
    _allowed_clients = ["App", "Website"]
    assert client in _allowed_clients
    return client


class UserCreationIn(BaseModel):
    name: str
    client: Annotated[str, AfterValidator(check_client)]
    bot_token: Optional[str] = None
    telegram_user_id: Optional[str] = None

    @model_validator(mode='after')
    def check_tg_token_userid(self) -> Any:
        if self.client == "App":
            if self.bot_token is None and self.telegram_user_id is None:
                raise ValueError('For Client Type "Application" You Need to provide Telegram Bot Token and '
                                 'Telegram'
                                 'User ID')
        return self


class UserCreationResponse(BaseModel):
    session_uuid: UUID


class UserGet(BaseModel):
    uuid: str
    name: str
    telegram_user_id: str


class UsersGetAll(BaseModel):
    sessions_available: List[UserGet]


class SSLCert(BaseModel):
    isSSLAvailable: bool
    SSLPullError: str
    Certificate: Optional[Dict]


class URLFeatures(BaseModel):
    length_url: int
    length_hostname: int
    ip: int
    nb_dots: int
    nb_qm: int
    nb_eq: int
    nb_slash: int
    nb_www: int
    ratio_digits_url: float
    ratio_digits_host: float
    tld_in_subdomain: int
    prefix_suffix: int
    shortest_word_host: int
    longest_words_raw: int
    longest_word_path: int
    phish_hints: int
    domain_age: int
    page_rank: float


class FakeDetectionResponse(BaseModel):
    original_url: str
    long_url: str
    ssl_cert: SSLCert
    features: URLFeatures
    host_name: str
    port: int
    ModelPrediction: str


class FakeDetectionIn(BaseModel):
    uuid: UUID
    text: str



import ssl
import socket
from socket import gaierror
import OpenSSL
from datetime import datetime
from ssl import SSLCertVerificationError, SSLError


def get_ssl_certificate(host, port=443, timeout=10) -> (bool, str):
    flag = True
    err = ""
    gai_flag = False
    try:
        context = ssl.create_default_context()
        conn = socket.create_connection((host, port))
        sock = context.wrap_socket(conn, server_hostname=host)
        flag = True
    except SSLCertVerificationError as e:
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        flag = False
        err = str(e)
    except SSLError as e:
        flag = False
        err = str(e)
        gai_flag = True
    except gaierror as e:
        flag = False
        err = str(e)
        gai_flag = True
    finally:
        if gai_flag:
            return flag, err, None
        conn = socket.create_connection((host, port))
        sock = context.wrap_socket(conn, server_hostname=host)
        sock.settimeout(timeout)
        der_cert = sock.getpeercert(True)
        _cert = ssl.DER_cert_to_PEM_cert(der_cert)
        x509 = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, _cert)
        subject = x509.get_subject().get_components()
        issuer = x509.get_issuer().get_components()
        result = {
            'subject': {key.decode('utf-8'): value.decode('utf-8') for key, value in subject},
            'issuer': {key.decode('utf-8'): value.decode('utf-8') for key, value in issuer},
            'serialNumber': x509.get_serial_number(),
            'version': x509.get_version(),
            'notBefore': datetime.strptime(x509.get_notBefore().decode('utf-8'), '%Y%m%d%H%M%SZ'),
            'notAfter': datetime.strptime(x509.get_notAfter().decode('utf-8'), '%Y%m%d%H%M%SZ'),
        }

        extensions = (x509.get_extension(i) for i in range(x509.get_extension_count()))
        extension_data = {e.get_short_name().decode('utf-8'): str(e) for e in extensions}
        result.update(extension_data)
        sock.close()
        return flag, err, result






from fastapi import HTTPException, Security, status
from fastapi.security.api_key import APIKeyQuery

from fakeurldetector import Config, LOGGER

api_key_query_auth = APIKeyQuery(name="api_key", auto_error=True)


async def verify_api_key(api_key_query: str = Security(api_key_query_auth)):
    if api_key_query != Config.API_KEY:
        LOGGER.debug("API Key Mismatch")
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Unauthorized",
        )



import os

from dotenv import load_dotenv

load_dotenv(".env")


class Config(object):
    OPEN_PAGE_RANK_API_KEY = os.environ.get("OPEN_PAGE_RANK_API_KEY", -1)
    DB_URL = os.environ.get("DB_URL", "sqlite:///./users.sqlite")
    API_KEY = os.environ.get("API_KEY", "test")



from uuid import UUID, uuid4

from fastapi import APIRouter
from fastapi.encoders import jsonable_encoder
from fastapi.responses import ORJSONResponse

from fakeurldetector import DATABASE, LOGGER
from ..helpers.database import users
from ..helpers.models import UserCreationIn, UserCreationResponse, UserGet, UsersGetAll

UsersRouter = APIRouter(
    prefix="/users",
    tags=["User Sessions"],
    responses={404: {"error": "Not found"}},
)


@UsersRouter.post(
    "/create", response_class=ORJSONResponse, response_model=UserCreationResponse
)
async def create_user(user_config: UserCreationIn):
    try:
        _uuid = uuid4()
        _q = users.insert().values(
            uuid=str(_uuid),
            name=user_config.name,
            client=user_config.client,
            bot_token=user_config.bot_token,
            telegram_user_id=user_config.telegram_user_id
        )
        await DATABASE.execute(_q)
        return ORJSONResponse(content={"user_uuid": _uuid})
    except Exception as e:
        print(e)


@UsersRouter.delete(
    "/delete/{uuid}",
    response_class=ORJSONResponse,
    response_model=UserCreationResponse,
)
async def delete_user(uuid: UUID):
    try:
        _q = users.delete().where(users.c.uuid == str(uuid))
        await DATABASE.execute(_q)
        return ORJSONResponse(content={"user_uuid": uuid})
    except Exception as e:
        print(e)


@UsersRouter.get(
    "/get/{uuid}", response_class=ORJSONResponse, response_model=UserGet
)
async def get_user(uuid: UUID):
    try:
        _session_q = users.select().where(users.c.uuid == str(uuid))
        _session = await DATABASE.fetch_one(_session_q)
        return ORJSONResponse(content=jsonable_encoder({_session}))
    except Exception as e:
        print(e)


@UsersRouter.get("/all", response_class=ORJSONResponse, response_model=UsersGetAll)
async def get_all_users():
    try:
        _all = users.select()
        _data = await DATABASE.fetch_all(_all)
        return ORJSONResponse(
            content={"sessions_available": [jsonable_encoder(data) for data in _data]}
        )
    except Exception as e:
        print(e)


@UsersRouter.delete("/all", response_class=ORJSONResponse)
async def delete_all_sessions():
    try:
        _q = users.delete()
        await DATABASE.execute(_q)
        return ORJSONResponse(content={"result": "Deleted all the sessions"})
    except Exception as e:
        print(e)


from .models import *
from .api_auth import verify_api_key
from .get_ssl_certificate import get_ssl_certificate
from .selenium_redirected_url import get_redirected_url
from .url_helper import extract_url_features


import os.path
from selenium import webdriver
from selenium.webdriver.chrome.service import Service as ChromeService
from selenium.webdriver.common.by import By
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as ec
from selenium.common.exceptions import WebDriverException
import platform


def get_chrome_bin_loc():
    return os.path.join(os.getcwd(), "binaries", "chrome", "chrome")


def get_chrome_options():
    chrome_options = Options()
    chrome_options.add_argument("--ignore-ssl-errors=yes")
    chrome_options.add_argument("--ignore-certificate-errors")
    chrome_options.AcceptInsecureCertificates = True
    chrome_options.add_argument("--headless")
    chrome_options.add_argument("--no-sandbox")
    chrome_options.accept_insecure_certs = True
    if platform.system().lower() == "linux":
        chrome_options.binary_location = get_chrome_bin_loc()
    return chrome_options


def get_redirected_url(url: str) -> (bool, str):
    chrome_service = ChromeService(executable_path=os.path.join(os.getcwd(), "binaries", "chromedriver"))
    driver = webdriver.Chrome(service=chrome_service, options=get_chrome_options())
    # Check for redirection
    try:
        driver.get(url)
        wait = WebDriverWait(driver, 10)
        wait.until(ec.presence_of_element_located((By.TAG_NAME, 'body')))
        cur_url = driver.current_url

        return True, cur_url
    except WebDriverException as e:
        return False, e.msg
    finally:
        driver.quit()



import whois
from datetime import datetime
import tldextract
from urllib.parse import urlparse
import urllib
import re
import requests
from fakeurldetector import Config


def get_domain_age(domain_name: str) -> int:
    try:
        # Perform a WHOIS lookup for the domain
        domain_info = whois.whois(domain_name)

        # Extract the creation date from the WHOIS information
        creation_date = domain_info.creation_date

        # Ensure that creation_date is a datetime object
        if isinstance(creation_date, list):
            creation_date = creation_date[0]

        # Calculate the domain age
        current_date = datetime.now()
        domain_age = current_date - creation_date
        return domain_age.days
    except Exception as e:
        # Return -1 if domain age is not available
        return -1


def page_rank(url: str) -> int:
    o = urllib.parse.urlsplit(url)
    url_to_check = o[1]

    # Make a GET request to the Open PageRank API
    url1 = 'https://openpagerank.com/api/v1.0/getPageRank'
    params = {
        'domains[]': url_to_check
    }
    headers = {
        'API-OPR': Config.OPEN_PAGE_RANK_API_KEY
    }

    response = requests.get(url1, params=params, headers=headers)
    data = response.json()

    return data['response'][0]['page_rank_decimal']


# Function to extract features from a URL
def extract_url_features(url):
    parsed_url = urlparse(url)

    # Initialize a dictionary to store the features
    features = {'length_url': len(url), 'length_hostname': len(parsed_url.netloc)}

    # Feature: IP Address (True if the hostname is an IP address, False otherwise)
    a1 = parsed_url.netloc.replace('.', '').isdigit()
    if a1 == "True":
        features["ip"] = 1
    else:
        features["ip"] = 0

    # Feature: Number of dots in the URL
    features['nb_dots'] = url.count('.')

    # Feature: Number of question marks in the URL
    features['nb_qm'] = url.count('?')

    # Feature: Number of equal signs in the URL
    features['nb_eq'] = url.count('=')

    # Feature: Number of slashes in the URL
    features['nb_slash'] = url.count('/')

    # Feature: Number of 'www' subdomains
    features['nb_www'] = parsed_url.netloc.count('www')

    # Feature: Ratio of digits in the URL
    try:
        features['ratio_digits_url'] = sum(c.isdigit() for c in url) / len(url)
    except Exception as e:
        features['ratio_digits_url'] = 0

    # Feature: Ratio of digits in the hostname
    try:
        features['ratio_digits_host'] = sum(c.isdigit() for c in parsed_url.netloc) / len(parsed_url.netloc)
    except Exception as e:
        features['ratio_digits_host'] = 0

    # Feature: TLD (Top-Level Domain) in the subdomain
    extracted_domain = tldextract.extract(url)
    domain = extracted_domain.domain + '.' + extracted_domain.suffix
    tld = extracted_domain.suffix
    subdomain = extracted_domain.subdomain
    if subdomain.count(tld) > 0:
        features['tld_in_subdomain'] = 1
    else:
        features['tld_in_subdomain'] = 0

    # Feature: Prefix and Suffix in the hostname
    if re.findall(r"https?://[^\-]+-[^\-]+/", url):
        features['prefix_suffix'] = 1
    else:
        features['prefix_suffix'] = 0

    # Feature: Length of the shortest word in the hostname
    words = parsed_url.netloc.split('.')
    features['shortest_word_host'] = min(len(word) for word in words)

    # Feature: Length of the longest words in the URL
    features['longest_words_raw'] = max(len(word) for word in url.split('/'))

    # Feature: Length of the longest word in the path
    path = parsed_url.path
    features['longest_word_path'] = max(len(word) for word in path.split('/'))

    # Feature: Phishing hints (You can implement a specific algorithm to detect phishing hints)
    _hints = ['wp', 'login', 'includes', 'admin', 'content', 'site', 'images', 'js', 'alibaba', 'css', 'myaccount',
              'dropbox', 'themes', 'plugins', 'signin', 'view']
    path = parsed_url.path
    c = 0
    for hint in _hints:
        c += path.lower().count(hint)
    features["phish_hints"] = c

    # Feature: Domain Age (You can implement a function to retrieve the domain age)

    features["domain_age"] = get_domain_age(parsed_url.netloc)

    # Feature: Page Rank (You can implement a function to retrieve the page rank)
    features["page_rank"] = float(page_rank(url))

    return features



from databases import Database
from fakeurldetector.data import MODEL_DIR
from logging.handlers import RotatingFileHandler
import logging

from .BaseConfig import Config
from .version import __version__

# Logger
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - %(message)s [%(filename)s:%(lineno)d]",
    datefmt="%d-%b-%y %H:%M:%S",
    handlers=[
        RotatingFileHandler("log.txt", maxBytes=50000000, backupCount=10),
        logging.StreamHandler(),
    ],
)
logging.getLogger("pyrogram").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
LOGGER = logging.getLogger(__name__)


DATABASE = Database(Config.DB_URL)



from setuptools import find_packages, setup

from pathlib import Path
from fakeurldetector.version import __version__


def get_requirements():
    requirements_list = []

    with Path("requirements.txt").open() as reqs:
        for install in reqs:
            if install.startswith("#"):
                continue
            requirements_list.append(install.strip())

    return requirements_list


README = Path("Readme.md").read_text()

setup(
    name='URLRakshak',
    version=__version__,
    packages=find_packages(),
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://github.com/ViswanathBalusu/SafeURLInspector',
    license='GPL3.0',
    author='viswanathbalusu',
    author_email='ckvbalusu@gmail.com',
    include_package_data=True,
    description='Fake URL Inspector',
    platforms="any",
    install_requires=get_requirements(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 5 - Production/Stable"
    ],
    package_data={
        "": ["data/model.pkl"],
    },
    python_requires=">=3.11",
    scripts=['scripts/URLRakshak'],
)



import time

from fastapi import FastAPI, Request, Security
from fastapi.middleware.cors import CORSMiddleware

from fakeurldetector import DATABASE, __version__

from .helpers.api_auth import verify_api_key
from .routers import *
from fakeurldetector import Config, LOGGER


FakeURLDetection = FastAPI(
    title="Fake URL Detection",
    version=__version__,
    docs_url="/docs",
    redoc_url=None,
    openapi_url="/fake_url.json",
    dependencies=[Security(verify_api_key)],
)


@FakeURLDetection.on_event("startup")
async def database_dir_init():
    if Config.OPEN_PAGE_RANK_API_KEY == -1 or Config.OPEN_PAGE_RANK_API_KEY is None:
        LOGGER.error("Cant Start the Program Without the Open Page Rank API Key")
        exit(1)
    await DATABASE.connect()


@FakeURLDetection.on_event("shutdown")
async def database_close():
    await DATABASE.disconnect()


#
@FakeURLDetection.middleware("http")
async def add_process_time_header(request: Request, call_next):
    start_time = time.time()
    response = await call_next(request)
    process_time = time.time() - start_time
    response.headers["X-Process-Time"] = str(round(process_time * 1000, 3)) + " ms"
    return response


FakeURLDetection.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
FakeURLDetection.include_router(UsersRouter)
FakeURLDetection.include_router(FakeDetectionRouter)



from uuid import UUID

from fastapi import APIRouter, HTTPException
from ..helpers.database import users
from fakeurldetector import DATABASE, LOGGER, MODEL_DIR
from fastapi.encoders import jsonable_encoder
from ..helpers import get_ssl_certificate, get_redirected_url, extract_url_features
from ..helpers.models import FakeDetectionResponse, FakeDetectionIn
from telebot.async_telebot import AsyncTeleBot
from telebot.util import quick_markup
from urllib.parse import urlparse
import pickle
import re
import pandas as pd

FakeDetectionRouter = APIRouter(
    prefix="/fake",
    tags=["Fake Detection API Endpoints"],
    responses={404: {"error": "Not found"}},
)


@FakeDetectionRouter.post(
    "/detect", response_model=FakeDetectionResponse
)
async def fake_detect(uuid_text: FakeDetectionIn):
    _session_q = users.select().where(users.c.uuid == str(uuid_text.uuid))
    _session = await DATABASE.fetch_one(_session_q)
    url_pattern = r"https?://\S+|www\.\S+|ftp://\S+|\S+\.\S+/\S+"
    uuid_text.text = re.findall(url_pattern, uuid_text.text)
    if uuid_text.text is None:
        raise HTTPException(status_code=404, detail={"Could not find URL in the text you sent"})
    uuid_text.text = uuid_text.text[0]
    _dict = {}
    if _session[2] == "App":
        initial_text = f"""
<b>Given URL :</b><code> {uuid_text.text}</code>    
<b>Checking if URL Redirect 🔄</b>
"""
        _bot = AsyncTeleBot(str(_session[3]))
        _msg = await _bot.send_message(int(_session[4]), initial_text, parse_mode="HTML")
    _dict["original_url"] = str(uuid_text.text)
    check, _original_url = get_redirected_url(str(uuid_text.text))
    if _original_url == "data:,":
        _original_url = uuid_text.text
    if not check:
        LOGGER.info(f"Selenium Could not load the Site 😞'{uuid_text.text}'")
        _original_url = uuid_text.text
    LOGGER.info(f"Redirect Found : '{uuid_text.text}'")
    _dict["long_url"] = _original_url
    if _session[2] == "App":
        if check:
            middle_text = f"""
<b>Given URL :</b><code> {uuid_text.text}</code>    
<b>Redirected URL :</b><code> {_original_url}</code>           
"""
        else:
            middle_text = f"""
<b>Given URL :</b><code> {uuid_text.text}</code>    
<b>Resolving Failed due to Error 😞 :</b><code> {_original_url}</code>         
"""
        await _bot.edit_message_text(middle_text, chat_id=int(_session[4]), message_id=_msg.id, parse_mode="HTML")

    if _session[2] == "App":
        ssl_msg = """
<b>Getting the SSL Details 🔒, Stay Tight!!</b>
"""
        await _bot.edit_message_text(middle_text + ssl_msg, chat_id=int(_session[4]), message_id=_msg.id,
                                     parse_mode="HTML")
    parsed_hostname = urlparse(_original_url).netloc
    parsed_port = urlparse(_original_url).port
    if parsed_port is None:
        parsed_port = 443
    check, err, cert = get_ssl_certificate(parsed_hostname, parsed_port, timeout=10)
    _dict["ssl_cert"] = {
        "isSSLAvailable": check,
        "SSLPullError": err,
        "Certificate": cert
    }
    if cert is not None:
        _dict["ssl_cert"]["Certificate"]["notBefore"] = _dict["ssl_cert"]["Certificate"]["notBefore"].strftime('%m/%d/%Y')
        _dict["ssl_cert"]["Certificate"]["notAfter"] = _dict["ssl_cert"]["Certificate"]["notAfter"].strftime('%m/%d/%Y')

    if _session[2] == "App":
        if check:
            middle_text += f"""
<b>SSL Certificate is Valid 🔒😃</b>

<b>Date of Issue 📅 : </b><code> {cert["notBefore"]}</code>
<b>Date of Expiry 📅 : </b><code> {cert["notAfter"]}</code>
<b>Issuer 🌐: </b><code> {cert["issuer"]}</code>
<b>Issued to 💼: </b><code> {cert["subject"]}</code>
"""
            await _bot.edit_message_text(middle_text, chat_id=int(_session[4]), message_id=_msg.id,
                                         parse_mode="HTML")
        else:
            if cert is not None:
                middle_text += f"""
<b>SSL Certificate is Invalid 😬</b>

<b>Error Message ❌: </b><code>{err}</code>
<b>Date of Issue 📅 : </b><code> {cert["notBefore"]}</code>
<b>Date of Expiry 📅 : </b><code> {cert["notAfter"]}</code>
<b>Issuer 🌐: </b><code> {cert["issuer"]['O']}</code>
<b>Issued to 💼: </b><code> {cert["subject"]['CN']}</code>              
"""
                await _bot.edit_message_text(middle_text, chat_id=int(_session[4]), message_id=_msg.id,
                                             parse_mode="HTML")
            else:
                middle_text += f"""
<b>Could Not pull the SSL Certificate 😕(Probably it never existed)</b>

<b>Error Message ❌: </b><code>{err}</code>
"""
                await _bot.edit_message_text(middle_text, chat_id=int(_session[4]), message_id=_msg.id,
                                             parse_mode="HTML")

    _features = extract_url_features(_original_url)
    if _session[2] == "App":
        middle_text += f"""
<b>Features of Long URL</b>
    
<b>Host Name 📡: </b> <code>{parsed_hostname}</code>
<b>Port 🔌: </b> <code>{parsed_port}</code>
<b>Length of URL 📏: </b> <code>{_features["length_url"]}</code>
<b>Is hostname an IP? 🌐: </b><code>{"True" if _features["ip"] == 1 else "False"}</code>
<b>Page Rank 📊: </b><code>{_features["page_rank"]}</code>
<b>Domain Age 🕰️: </b><code>{"No Data" if _features["domain_age"] == -1 else _features["domain_age"]} Days</code>
"""
        await _bot.edit_message_text(middle_text, chat_id=int(_session[4]), message_id=_msg.id,
                                     parse_mode="HTML")
        ai_model_message = """
<b>Now Passing all the parameters to AI Model 🤖</b>
"""
        await _bot.edit_message_text(middle_text + ai_model_message, chat_id=int(_session[4]), message_id=_msg.id,
                                     parse_mode="HTML")
    _dict["features"] = _features
    _dict["host_name"] = parsed_hostname
    _dict["port"] = parsed_port
    X = pd.DataFrame(_features, index=[0]).iloc[:, :].values
    with open(MODEL_DIR, 'rb') as model:
        _model = pickle.load(model)
    _pred = _model.predict(X)

    if _session[2] == "App":
        markup = quick_markup({f'Model Prediction 🤖: {"Legit Site✅" if _pred == 0 else "Fake❌"}': {'url': _original_url}})
        await _bot.edit_message_text(middle_text, chat_id=int(_session[4]), reply_markup=markup,message_id=_msg.id,
                                     parse_mode="HTML")
    _dict["ModelPrediction"] = "Legit Site" if _pred == 0 else "Fake"
    return jsonable_encoder(_dict)
